import os
import re
import sys
from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Optional

from .utils import to_glob, upath


def ll(
    *paths: str,
    depth: int = 1,
    unit_glob: Optional[str] = None,
    type: Optional[str] = None,
    glob: str = "**/*",
    regex: str = r"^(?!.*(\.git\/|__pycache__\/|\.ipynb_checkpoints\/|\.pytest_cache\/|\.vscode\/|\.idea\/|\.DS_Store)).*$",
    number_limit: int = 10,
    sort_paths: str = "asc",
    info: bool = False,
    extension_most_common: int = 1,
) -> None:
    """Compute statistics for the directory recursively."""

    paths = paths or (".",)

    unit_glob = unit_glob or (to_glob(depth) if depth > 0 else "**/*")

    stat_ls = []

    for path in paths:
        _path = Path(upath(path))

        if type:
            dir_path_ls = [
                p
                for p in (_path.glob(unit_glob))
                if getattr(p, "is_" + type.replace("f", "file").replace("d", "dir"))()
            ]
        else:
            dir_path_ls = list(_path.glob(unit_glob))

        if sort_paths:
            assert isinstance(sort_paths, str), sort_paths
            dir_path_ls.sort(reverse=(sort_paths.lower().startswith("desc")))

        if info:
            for dir_path in dir_path_ls:
                d = _get_stat(
                    dir_path=dir_path,
                    glob=glob,
                    regex=regex,
                    type=type,
                    number_limit=number_limit,
                    extension_most_common=extension_most_common,
                )
                if d:
                    stat_ls.append(d)
        else:
            sys.stdout.write(
                "\n".join(
                    [str(p) + (os.sep if p.is_dir() else "") for p in dir_path_ls]
                )
                + "\n"
            )

    if info:
        _output_stat(stat_ls)


def _get_stat(
    dir_path: Path,
    glob: str,
    regex: Optional[str],
    type: Optional[str],
    number_limit: int,
    extension_most_common: int,
) -> Optional[Dict]:
    rx = re.compile(regex) if regex else None
    if dir_path.is_dir():
        try:
            path_ls = [p for p in dir_path.glob(glob) if p.is_file()]
            if rx:
                path_ls = [p for p in path_ls if rx.match(str(p))]
            num_files = len(path_ls)
            _num_files = f"{num_files:,}"
        except PermissionError:
            return None
    else:
        path_ls = [dir_path]
        num_files = len(path_ls)
        _num_files = ""

    if not path_ls:
        return None

    divisor = None
    if num_files > number_limit:
        divisor = num_files / number_limit

    path_ls = path_ls[:number_limit]

    size_ls = []
    # mtime_ls = []
    ext_ls = []
    for p in path_ls:
        try:
            st = p.stat()
        except Exception:
            continue
        size_ls.append(st.st_size)
        # mtime_ls.append(st.st_mtime)

        ext = p.suffix
        ext_ls.append(ext)

    total_size = sum(size_ls)
    max_size = max(size_ls)
    # latest_mtime = max(mtime_ls)

    common_ext_dict = {}
    if extension_most_common:
        common_ext_count_ls = Counter(ext_ls).most_common(extension_most_common)
        common_ext_dict = {
            "ext_" + str(i + 1): common_ext_count_ls[i][0]
            for i in range(extension_most_common)
        }

    d = {"path": str(dir_path) + (os.sep if dir_path.is_dir() else "")}

    if divisor:
        total_size = round(total_size * divisor)
        _total_size = f"~ {total_size:,}"
        _max_size = f">= {max_size:,}"
    else:
        _total_size = f"{total_size:,}"
        _max_size = f"{max_size:,}"

    if type == "file":
        d.update({"size": _total_size})
    else:
        d.update(
            {
                "files": _num_files,
                "total_size": _total_size,
                "max_size": _max_size,
                # "latest_mtime": latest_mtime,
            },
        )
        d.update(common_ext_dict)

    return d


def _output_stat(stat_ls: List[Dict[str, Any]]) -> None:
    try:
        import pandas as pd

        colalign_ls = None
        if stat_ls:
            colalign_ls = ["right"] * len(stat_ls[0])
            colalign_ls[0] = "left"

        df = pd.DataFrame(stat_ls)
        md = df.to_markdown(index=False, colalign=colalign_ls)
        sys.stdout.write(str(md) + "\n")
    except Exception:
        import json

        sys.stdout.write(json.dumps(stat_ls, indent=2) + "\n")



from pathlib import Path
from typing import Any, List, Optional

from .utils import infer_format, stdout_lines

DATA_TYPES = {
    "csv",
    "ipc",
    "parquet",
    "database",
    "json",
    "ndjson",
    "avro",
    "excel",
    "delta",
}

TRUE_VALUES = {"", "True", "true", "T", "t", "1"}


def polars(
    *paths: str,
    input_format: Optional[str] = None,
    columns: Optional[List[str]] = None,
    excluding_columns: Optional[List[str]] = None,
    filepath_column: Optional[str] = None,
    streaming: Optional[str] = None,  # actually bool
    fetch: Optional[int] = None,
    join: Optional[str] = None,
    on: Optional[str] = None,
    left_on: Optional[str] = None,
    right_on: Optional[str] = None,
    suffix: str = "_right",
    validate: str = "m:m",
    head: Optional[int] = None,
    tail: Optional[int] = None,
    sample: Optional[int] = None,
    method: Optional[str] = None,
    output_format: Optional[str] = None,
    output_path: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Read and transform tabular files using polars."""

    """Workaround for unexpected behavior of Fire"""
    kwargs.pop("input_format", None)
    kwargs.pop("columns", None)
    kwargs.pop("excluding_columns", None)
    kwargs.pop("filepath_column", None)
    kwargs.pop("streaming", None)
    kwargs.pop("fetch", None)
    kwargs.pop("join", None)
    kwargs.pop("on", None)
    kwargs.pop("left_on", None)
    kwargs.pop("right_on", None)
    kwargs.pop("suffix", "_right")
    kwargs.pop("validate", "m:m")
    kwargs.pop("head", None)
    kwargs.pop("tail", None)
    kwargs.pop("sample", None)
    kwargs.pop("method", None)
    kwargs.pop("output_format", None)
    kwargs.pop("output_path", None)

    _output_format = (
        infer_format(
            output_format, output_path, DATA_TYPES.union({"markdown"}), polars=True
        )
        or "csv"
    )

    streaming_flag = streaming in TRUE_VALUES

    import polars as pl

    def activate(
        df: pl.LazyFrame,
        fetch: Optional[int] = None,
        streaming: bool = False,
    ) -> pl.DataFrame:
        return (
            df.fetch(n_rows=fetch, streaming=streaming)
            if fetch
            else df.collect(streaming=streaming)
        )

    ls = []
    for path in paths:
        _input_format = infer_format(input_format, path, DATA_TYPES, polars=True)
        if not _input_format:
            continue

        _kwargs = kwargs.copy()
        if input_format == "csv":
            _kwargs.setdefault("missing_utf8_is_empty_string", True)
            _kwargs.setdefault("infer_schema_length", 0)

        read_func = getattr(pl, "scan_" + _input_format, None)
        if read_func is None:
            read_func = getattr(pl, "read_" + _input_format)
            df = read_func(path, **_kwargs).lazy()
        else:
            df = read_func(path, **_kwargs)

        if columns:
            df = df.select(columns)
        if excluding_columns:
            if isinstance(excluding_columns, str):
                excluding_columns = [excluding_columns]
            _columns = df.columns
            _columns = [c for c in _columns if c not in excluding_columns]
            df = df.select(_columns)

        if filepath_column:
            df = df.with_columns(pl.lit(path).alias(filepath_column))

        ls.append(df)

    if not ls:
        return
    if len(ls) == 1:
        df = ls[0]
    elif join is not None:
        df = ls[0]
        for right_df in ls[1:]:
            df = df.join(
                right_df,
                on=on,
                how=join,
                left_on=left_on,
                right_on=right_on,
                suffix=suffix,
                validate=validate,
            )
    else:
        df = pl.concat(ls)

    subset_ls = []
    if head is not None:
        subset_ls.append(df.head(head))
    if tail is not None:
        subset_ls.append(df.tail(tail))
    if subset_ls:
        df = pl.concat(subset_ls)

    if sample is not None:
        df = activate(df, fetch, streaming_flag)
        df = df.sample(sample)
        df = df.lazy()

    if method is not None:
        df = eval("df." + method)

    df = activate(df, fetch, streaming_flag)

    if not isinstance(df, pl.DataFrame):
        text = f"{df}"
        if output_path:
            Path(output_path).write_text(text)
        else:
            stdout_lines(text)
        return

    if _output_format == "markdown":

        def write_func(output_path: Optional[str] = None) -> Optional[str]:
            from io import StringIO

            import pandas as pd

            nonlocal df
            csv_text = df.write_csv()
            df = pd.read_csv(StringIO(csv_text), dtype=str, keep_default_na=False)
            return df.to_markdown(output_path, index=False)

    else:
        write_func = getattr(df, "write_" + _output_format)

    if output_path:
        write_func(output_path)
    else:
        print(df)



import re
import sys
from pathlib import Path
from typing import Any, List, Tuple

from .utils import (
    get_exception_msg,
    subprocess_run,
    subprocess_run_stdout,
    to_glob,
    upath,
)


def recur(
    path: str,
    *scripts: str,
    **kwargs: Any,
) -> Tuple[List[str], List[str], str, bool]:
    avoid_fd = kwargs.pop("avoid_fd", None)
    type = kwargs.pop("type", "f")
    depth = kwargs.pop("depth", None)
    glob = kwargs.pop("glob", None)
    absolute_path = kwargs.pop("absolute_path", False)
    regex = kwargs.pop(
        "regex",
        r"^(?!.*(\.git\/|__pycache__\/|\.ipynb_checkpoints\/|\.pytest_cache\/|\.vscode\/|\.idea\/|\.DS_Store)).*$",
    )
    sort_paths = kwargs.pop("sort_paths", "asc")
    replace_str = kwargs.pop("replace_str", "@@")
    show_paths = kwargs.pop("show_paths", False)
    show_scripts = kwargs.pop("show_scripts", False)

    script_ls = list(scripts)
    if len(kwargs) and len(script_ls) == 1:
        script_ls = script_ls[0].split(" ")
    for k, v in kwargs.items():
        if isinstance(v, bool) and not v:
            continue
        if len(k) >= 2:
            script_ls.append("--" + k)
        elif len(k) == 1:
            script_ls.append("-" + k)
        else:
            raise NotImplementedError
        if isinstance(v, bool):
            continue
        script_ls.append(str(v))

    if not script_ls:
        script_ls = ["echo"]

    if replace_str and all(replace_str not in script for script in script_ls):
        if len(script_ls) >= 2:
            script_ls.append(replace_str)
        else:
            script_ls[0] += " " + replace_str

    _path = Path(upath(path))
    assert _path.exists(), str(_path.resolve()) + " does not exist."

    if _path.is_file():
        path_ls = [str(_path)]
    else:
        if avoid_fd:
            glob = glob or (to_glob(depth) if depth else "**/*")
            path_ls = [
                str(p)
                for p in _path.glob(glob)
                if (not type)
                or getattr(p, "is_" + type.replace("f", "file").replace("d", "dir"))()
            ]
        else:
            path_ls = (
                subprocess_run_stdout(
                    "fd "
                    + ("" if not absolute_path else f"--absolute-path ")
                    + ("" if not type else f"--type {type} ")
                    + ("" if depth is None else f"--max-depth {depth} ")
                    + ("''" if glob is None else f"--glob {glob} ")
                    + f"{path}",
                    verbose=False,
                )
                .strip()
                .split("\n")
            )
        if regex:
            rx = re.compile(regex)
            path_ls = [p for p in path_ls if rx.match(p)]
        if sort_paths:
            assert isinstance(sort_paths, str), sort_paths
            path_ls.sort(reverse=(sort_paths.lower().startswith("desc")))

    if show_paths:
        sys.stdout.write(
            "[Searching files]" + str("\n".join(["    " + p for p in path_ls]) + "\n")
        )
    return path_ls, script_ls, replace_str, show_scripts


def under(
    path: str,
    *scripts: str,
    **kwargs: str,
) -> None:
    """Run any scripts for each file under a directory recursively."""

    path_ls, script_ls, replace_str, show_scripts = recur(
        path,
        *scripts,
        **kwargs,
    )

    for p in path_ls:
        try:
            running_script_ls = script_ls
            if replace_str:
                running_script_ls = [
                    script.replace(replace_str, p)
                    if isinstance(script, str)
                    else script
                    for script in script_ls
                ]

            if len(running_script_ls) == 1:
                subprocess_run(running_script_ls[0], show_scripts)
            else:
                subprocess_run(running_script_ls, show_scripts)
        except Exception:
            msg = get_exception_msg()
            sys.stdout.write(msg)
            continue


def batch(
    path: str,
    *scripts: str,
    **kwargs: str,
) -> None:
    """Run any scripts for a batch of files in a directory recursively."""

    path_ls, script_ls, replace_str, show_scripts = recur(
        path,
        *scripts,
        **kwargs,
    )

    try:
        if len(script_ls) == 1:
            running_script = script_ls[0]
            if replace_str:
                running_script = running_script.replace(replace_str, " ".join(path_ls))
            subprocess_run(running_script, show_scripts)
        else:
            running_script_ls = []
            for script in script_ls:
                if replace_str and (script == replace_str):
                    running_script_ls.extend(path_ls)
                else:
                    running_script_ls.append(script)

            subprocess_run(running_script_ls, show_scripts)
    except Exception:
        msg = get_exception_msg()
        sys.stdout.write(msg)



import fire

from .ll import ll
from .pandas import pandas
from .polars import polars
from .recur import batch, under
from .search import find, search


def main() -> None:
    fire.Fire(
        {
            "ll": ll,
            "pandas": pandas,
            "polars": polars,
            "batch": batch,
            "under": under,
            "find": find,
            "search": search,
        },
    )


def xpandas() -> None:
    fire.Fire(pandas)


def xpolars() -> None:
    fire.Fire(polars)


def xbatch() -> None:
    fire.Fire(batch)


def xunder() -> None:
    fire.Fire(under)


def xfind() -> None:
    fire.Fire(find)


def xsearch() -> None:
    fire.Fire(search)


def xll() -> None:
    fire.Fire(ll)



from pathlib import Path
from typing import Any, List, Optional, Tuple

from .utils import infer_format, stdout_lines

DATA_TYPES = {
    "pickle",
    "table",
    "csv",
    "fwf",
    "clipboard",
    "excel",
    "json",
    "html",
    "xml",
    "hdf",
    "feather",
    "parquet",
    "orc",
    "sas",
    "spss",
    "sql_table",
    "sql_query",
    "sql",
    "gbq",
    "stata",
}


def pandas(
    *paths: str,
    package: str = "pandas",
    input_format: Optional[str] = None,
    columns: Optional[List[str]] = None,
    excluding_columns: Optional[List[str]] = None,
    filepath_column: Optional[str] = None,
    join: Optional[str] = None,
    merge: Optional[str] = None,
    on: Optional[str] = None,
    left_on: Optional[str] = None,
    right_on: Optional[str] = None,
    left_index: bool = False,
    right_index: bool = False,
    sort: bool = False,
    suffixes: Tuple[str, str] = ("_x", "_y"),
    copy: bool = True,
    indicator: bool = False,
    validate: Optional[str] = None,
    lsuffix: Optional[str] = None,
    rsuffix: Optional[str] = None,
    query: Optional[str] = None,
    head: Optional[int] = None,
    tail: Optional[int] = None,
    sample: Optional[int] = None,
    method: Optional[str] = None,
    output_format: Optional[str] = None,
    output_path: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Read and transform tabular files using pandas."""
    """Workaround for unexpected behavior of Fire"""
    kwargs.pop("package", None)
    kwargs.pop("input_format", None)
    kwargs.pop("columns", None)
    kwargs.pop("excluding_columns", None)
    kwargs.pop("filepath_column", None)
    kwargs.pop("join", None)
    kwargs.pop("merge", None)
    kwargs.pop("on", None)
    kwargs.pop("left_on", None)
    kwargs.pop("right_on", None)
    kwargs.pop("left_index", False)
    kwargs.pop("right_index", False)
    kwargs.pop("sort", False)
    kwargs.pop("suffixes", ("_x", "_y"))
    kwargs.pop("copy", True)
    kwargs.pop("indicator", False)
    kwargs.pop("validate", None)
    kwargs.pop("lsuffix", "")
    kwargs.pop("rsuffix", "")
    kwargs.pop("query", None)
    kwargs.pop("head", None)
    kwargs.pop("tail", None)
    kwargs.pop("sample", None)
    kwargs.pop("method", None)
    kwargs.pop("output_format", None)
    kwargs.pop("output_path", None)

    _output_format = (
        infer_format(output_format, output_path, DATA_TYPES.union({"markdown"}))
        or "csv"
    )

    if package == "modin":
        import modin.pandas as pd
    elif package == "pandas":
        import pandas as pd
    else:
        raise NotImplementedError(
            "'" + package + "' not supported. Set one of ['pandas', 'modin']"
        )
    import numpy as np  # noqa: F401

    if columns and isinstance(columns, str):
        columns = [columns]
    if excluding_columns and isinstance(excluding_columns, str):
        excluding_columns = [excluding_columns]

    ls = []
    for path in paths:
        _input_format = infer_format(input_format, path, DATA_TYPES)
        if not _input_format:
            continue
        read_func = getattr(pd, "read_" + _input_format)
        _kwargs = kwargs.copy()
        if input_format == "csv":
            _kwargs.setdefault("dtype", str)
            _kwargs.setdefault("keep_default_na", False)
            if columns:
                _kwargs.setdefault("usecols", columns)
        df = read_func(path, **_kwargs)

        if columns:
            df = df[columns]
        if excluding_columns:
            _columns = df.columns
            _columns = [c for c in _columns if c not in excluding_columns]
            df = df[_columns]

        if filepath_column:
            df[filepath_column] = path

        if query:
            df = df.query(query)
        ls.append(df)

    if not ls:
        return
    if len(ls) == 1:
        df = ls[0]
    elif merge is not None:
        df = ls[0]
        for right_df in ls[1:]:
            if merge == "anti":
                cols = df.columns
                df = (
                    df.reset_index()
                    .merge(
                        right_df,
                        on=on,
                        how="left",
                        left_on=left_on,
                        right_on=right_on,
                        left_index=left_index,
                        right_index=right_index,
                        sort=sort,
                        suffixes=suffixes,
                        copy=copy,
                        indicator=True,
                        validate=validate,
                    )
                    .set_index("index")
                )
                df = df.query('_merge == "left_only"')[cols]
            else:
                df = (
                    df.reset_index()
                    .merge(
                        right_df,
                        on=on,
                        how=merge,
                        left_on=left_on,
                        right_on=right_on,
                        left_index=left_index,
                        right_index=right_index,
                        sort=sort,
                        suffixes=suffixes,
                        copy=copy,
                        indicator=indicator,
                        validate=validate,
                    )
                    .set_index("index")
                )
    elif join is not None:
        df = ls[0]
        for right_df in ls[1:]:
            df = df.join(
                right_df,
                on=on,
                how=join,
                lsuffix=lsuffix,
                rsuffix=rsuffix,
                sort=sort,
                validate=validate,
            )
    else:
        df = pd.concat(ls, ignore_index=True)

    subset_ls = []
    if head is not None:
        subset_ls.append(df.head(head))
    if tail is not None:
        subset_ls.append(df.tail(tail))
    if subset_ls:
        df = pd.concat(subset_ls, ignore_index=True)

    if sample is not None:
        df = df.sample(sample)

    if method is not None:
        df = eval("df." + method)

    if not isinstance(df, pd.DataFrame):
        text = f"{df}"
        if output_path:
            Path(output_path).write_text(text)
        else:
            stdout_lines(text)
        return

    _write_func = getattr(df, "to_" + _output_format)

    def write_func(
        output_path: Optional[str] = None,
        index: bool = False,
    ) -> Optional[str]:
        nonlocal _write_func, df
        try:
            if _output_format == "json":
                import json

                ls = df.to_dict(orient="records")
                if output_path:
                    with open(output_path, "w") as f:
                        json.dump(ls, f)
                else:
                    return json.dumps(ls)
            return _write_func(output_path, index=index)
        except Exception:
            return _write_func(output_path)

    if output_path:
        write_func(output_path=output_path, index=False)
    else:
        print(df)



import subprocess
import sys
import traceback
from pathlib import Path
from time import sleep
from typing import List, Optional, Set, Union


def get_exception_msg() -> str:
    return "".join(traceback.format_exception(*sys.exc_info()))


def upath(
    path: str,
) -> str:
    if isinstance(path, str) and path.startswith("~"):
        path = str(Path.home()) + path[1:]
    if isinstance(path, str) and path.startswith("."):
        path = str(Path.cwd()) + path[1:]
    return path


def subprocess_run(
    script: Union[str, List[str]],
    verbose: bool = True,
) -> subprocess.CompletedProcess:
    if verbose:
        sys.stdout.write(r">>> " + str(script) + "\n")

        sleep(0.2)

    assert script, script
    shell = isinstance(script, str)
    return subprocess.run(script, shell=shell)


def subprocess_run_stdout(
    script: Union[str, List[str]],
    verbose: bool = True,
) -> str:
    if verbose:
        sys.stdout.write(r">>> " + str(script) + "\n")

        sleep(0.2)

    assert script, script
    shell = isinstance(script, str)
    result = subprocess.run(script, shell=shell, stdout=subprocess.PIPE)
    return result.stdout.decode()


def stdout_lines(text: Optional[str]) -> None:
    if not text:
        return None
    if not text.endswith("\n"):
        text += "\n"
    sys.stdout.write(text)
    return None


def infer_format(
    format: Optional[str],
    path: Optional[str],
    supported_types: Set[str],
    polars: bool = False,
) -> Union[str, None]:
    if isinstance(format, str) and format:
        _format = format
        _format = _format.replace("md", "markdown")
        if polars:
            _format = _format.replace("jsonl", "ndjson")
        assert _format in supported_types, (
            _format + " not in supported set: " + str(supported_types)
        )
        return _format
    if isinstance(path, str) and path:
        _format = path.split(".")[-1]
        _format = _format.replace("md", "markdown")
        if polars:
            _format = _format.replace("jsonl", "ndjson")
        if _format in supported_types:
            return _format
    return None


def to_glob(depth: Optional[int] = None) -> Optional[str]:
    if depth:
        return "".join(["*/"] * (depth - 1)) + "*"
    return None



import sys
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union

from .utils import subprocess_run

from importlib.util import find_spec


def run_search(
    text: str,
    target: Union[str, List, Tuple, Set],
    path: Path,
    sub: Optional[Union[str, List, Tuple, Set]] = None,
    wildcard: str = "*",
    separator: str = "/",
    verbose: int = 1,
    context: Optional[int] = None,
    plain: bool = False,
) -> str:
    assert isinstance(separator, str), str(separator) + ": " + str(type(separator))

    if isinstance(target, (list, set, tuple)):
        targets = target
    else:
        assert isinstance(target, str), str(type(target))
        targets = target.split(separator) if separator else [target]

    if isinstance(sub, (list, tuple, set)):
        subs = sub
    elif sub is None:
        subs = [None] * len(targets)
    else:
        assert isinstance(sub, str), str(type(sub))
        subs = sub.split(separator) if separator else [sub]

    assert len(targets) == len(subs), str(len(targets)) + " != " + str(len(subs))

    for _target, _sub in zip(targets, subs):
        target_ls = eval("'''" + _target + "'''").split(wildcard)

        replacing_ls = []
        end_index = 0

        while True:
            index = end_index
            start_index = None

            for target_ss in target_ls:
                index = text.find(target_ss, index)
                if index >= 0:
                    start_index = start_index or index
                    index = index + len(target_ss)
                else:
                    break
            if start_index and (index >= 0):
                end_index = index
                if _sub is not None:
                    replacing = text[start_index:end_index]
                    replacing_ls.append(replacing)
                if verbose >= 1:
                    if context is not None:
                        n_lines = context + 1
                        pre_context = "\n".join(
                            text[0:start_index].rsplit("\n", n_lines)[-n_lines:]
                        )
                        post_context = "\n".join(
                            text[end_index:].split("\n", n_lines)[:n_lines]
                        )
                    else:
                        pre_context = ""
                        post_context = ""
                    title = f">> {path} [{start_index}:{end_index}]"
                    target = text[start_index:end_index]
                    if (not find_spec("rich")) or plain:
                        print(title)
                        print(pre_context, end="")
                        print(target, end="")
                        print(post_context)
                    else:
                        from rich.console import Console

                        rc = Console()

                        rc.print(title)
                        rc.print(pre_context, end="")
                        rc.print(target, style="bold red", end="")
                        rc.print(post_context)
            else:
                break
        for replacing in list(set(replacing_ls)):
            text = text.replace(replacing, _sub)

    return text


def search(
    target: str,
    path: str,
    sub: Optional[str] = None,
    wildcard: str = "*",
    separator: str = "/",
    verbose: int = 1,
    context: Optional[int] = 1,
    plain: bool = False,
) -> None:
    """Search a keyword, which may include wildcards, in the text file content, and optionally substitute (replace)."""
    target = str(target)
    path = str(path)
    sub = str(sub) if sub is not None else sub

    _path = Path(path)
    try:
        text = _path.read_text()
    except Exception:
        if verbose >= 3:
            raise
        return

    text = run_search(
        text=text,
        target=target,
        path=_path,
        sub=sub,
        wildcard=wildcard,
        separator=separator,
        verbose=verbose,
        context=context,
        plain=plain,
    )

    if sub is not None:
        _path.write_text(text)


def find(
    target: str,
    path: str,
    sub: Optional[str] = None,
    wildcard: str = "*",
    separator: str = "/",
    verbose: int = 1,
    plain: bool = False,
) -> None:
    """Find a keyword, which may include wildcards, in the file path, and optionally substitute (replace)."""
    target = str(target)
    path = str(path)
    sub = str(sub) if sub is not None else sub

    text = path
    _path = Path(path)

    text = run_search(
        text=text,
        target=target,
        path=_path,
        sub=sub,
        wildcard=wildcard,
        separator=separator,
        verbose=verbose,
        plain=plain,
    )

    if (sub is not None) and _path.exists() and (text != path):
        sub_path = Path(text)
        sub_path.parent.mkdir(parents=True, exist_ok=True)
        subprocess_run(["mv", str(_path), str(sub_path)], verbose=False)



from setuptools import find_packages, setup

console_scripts = [
    "recurtx = recurtx.__main__:main",
    "xpandas = recurtx.__main__:xpandas",
    "xpolars = recurtx.__main__:xpolars",
    "xunder = recurtx.__main__:xunder",
    "xbatch = recurtx.__main__:xbatch",
    "xsearch = recurtx.__main__:xsearch",
    "xfind = recurtx.__main__:xfind",
    "xll = recurtx.__main__:xll",
]


with open("requirements.txt", encoding="utf-8") as f:
    requires = []
    for line in f:
        req = line.split("#", 1)[0].strip()
        if req and not req.startswith("--"):
            requires.append(req)

long_description = """
Please see:
https://github.com/Minyus/recurtx
"""

setup(
    name="recurtx",
    version="0.0.17",
    packages=find_packages(exclude=["tests"]),
    entry_points={"console_scripts": console_scripts},
    install_requires=requires,
    description="CLI to transform text files recursively",
    license="Apache Software License (Apache 2.0)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Minyus/recurtx",
    author="Yusuke Minami",
    author_email="me@minyus.github.com",
    zip_safe=False,
    keywords="CLI, recursive, text, find, xargs, sed",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
)
